from typing import List

import numpy as np
from matplotlib import pyplot as plt

from arc.plot import plot_grid

# 2D ndarray of integers, both dimensions from 1 to 30, values from 0 to 9 (integers)
ArcGrid = np.ndarray


def verify_is_arc_grid(grid: np.ndarray):
    assert (
        len(grid.shape) == 2
    ), f"Expected two-dimensional array, got array with shape {grid.shape}."
    assert issubclass(grid.dtype.type, np.integer), f"Expected an integer array."
    assert 0 <= np.amin(grid) <= np.amax(grid) <= 9, (
        f"ARC grid can only contain values between 0 and 9. "
        f"Found interval: [{np.amin(grid)}, {np.amax(grid)}]"
    )


# on ARC tasks, we are allowed to make multiple predictions,
# aiming for at least one of them to be correct.
ArcPrediction = List[ArcGrid]


class ArcIOPair:
    def __init__(self, x: ArcGrid, y: ArcGrid):
        self.x = x
        self.y = y

    def plot(self, show=True, title:str=None, fig_axes = None):

        fig, (ax1, ax2) = fig_axes or plt.subplots(1, 2)
        if title:
            fig.suptitle(title, fontsize=16)
        plot_grid(ax1, self.x)
        ax1.set_title("Input")
        plot_grid(ax2, self.y)
        ax2.set_title("Output")

        if show:
            plt.show()

    def __eq__(self, other):
        if not isinstance(other, ArcIOPair):
            return False
        return self.x == other.x and self.y == other.y


class ArcProblem:
    """A problem in ARC Challenge.

    ARC problem has uid (name of original json file),
    a list of demonstration pairs and a list of test pairs.
    """

    def __init__(
        self, uid: str, demo_pairs: List[ArcIOPair], test_pairs: List[ArcIOPair]
    ):
        self.uid = uid
        self.train_pairs = demo_pairs
        self.test_pairs = test_pairs

    @property
    def test_inputs(self):
        return [p.x for p in self.test_pairs]

    @property
    def test_outputs(self):
        return [p.y for p in self.test_pairs]



import pytest
import numpy as np

from arc.types import verify_is_arc_grid


def test_black_passes():
    background = np.zeros((3,3), np.uint8)

    verify_is_arc_grid(background)
    verify_is_arc_grid(background.astype(np.int64))


def test_3d_fails():
    cube = np.zeros((3, 3, 3), np.uint8)

    with pytest.raises(AssertionError):
        verify_is_arc_grid(cube)


def test_big_values_fail():
    grid = 111 * np.ones((3, 3), np.uint8)

    with pytest.raises(AssertionError):
        verify_is_arc_grid(grid)


def test_float_fail():
    grid =  np.random.normal(size=(3, 3))

    with pytest.raises(AssertionError):
        verify_is_arc_grid(grid)



from .data import train_set, eval_set, describe_task_set
from .types import ArcIOPair, ArcProblem
from .consts import ArcColors, colors_rgb
from . import agents

_train_probs_by_id = {p.uid : p for p in train_set}
_valid_probs_by_id = {p.uid : p for p in train_set}

def get_train_problem_by_uid(uid: str) -> ArcProblem:
    return _train_probs_by_id[uid]

def get_validation_problem_by_uid(uid: str) -> ArcProblem:
    return _valid_probs_by_id[uid]



from typing import List, Dict

from arc.data import eval_set
from arc.types import ArcProblem, ArcPrediction
from arc.agents import ArcAgent
from fractions import Fraction
import numpy as np

MAX_ATTEMPTS = 3


class ArcEvaluationResult:
    def __init__(self, raw_answers: Dict[ArcProblem, ArcPrediction] = None):
        self.raw_answers = raw_answers or {}
        self.correct = set()
        self.partially_correct = set()
        self.correct_shape: Dict[ArcProblem, Fraction] = {}
        self.accuracy = None
        self.accuracy_any = None
        self.shape_accuracy = None

    def add_answer(self, prob: ArcProblem, pred: List[ArcPrediction]) -> None:
        assert len(pred) == len(prob.test_pairs), (
            f"The problem has {len(prob.test_pairs)} tests, "
            f"but the agent returned {len(pred)} predictions."
        )
        self.raw_answers[prob] = pred
        tests_successful = []
        right_shapes = []
        for answer_grid, solution in zip(prob.test_outputs, pred):
            success = any(np.all(s == answer_grid) for s in solution[:MAX_ATTEMPTS])
            tests_successful.append(success)
            right_shapes.append(
                Fraction(
                    sum(answer_grid.shape == s.shape for s in solution[:MAX_ATTEMPTS]),
                    len(solution[:MAX_ATTEMPTS]),
                )
            )

        self.correct_shape[prob] = sum(right_shapes) / len(prob.test_pairs)
        if all(tests_successful):
            self.correct.add(prob)
        elif any(tests_successful):
            self.partially_correct.add(prob)

        self.accuracy = None
        self.accuracy_any = None
        self.shape_accuracy = None

    def compute_scores(self):
        N = len(self.raw_answers)
        self.accuracy = len(self.correct) / N
        self.accuracy_any = self.accuracy + len(self.partially_correct) / N
        self.shape_accuracy = sum(self.correct_shape.values()) / N

    def __repr__(self):
        return (
            f"ARC results for {len(self.raw_answers)} problems. Stats:"
            f"\n{'Accuracy':<25}: {self.accuracy:.1%}"
            f"\n{'Accuracy(at least one)':<25}: {self.accuracy_any:.1%}"
            f"\n{'Correct answer shape':<25}: {float(self.shape_accuracy):.1%}"
        )


def evaluate_agent(
    agent: ArcAgent, problems: List[ArcProblem] = eval_set
) -> ArcEvaluationResult:
    result = ArcEvaluationResult()
    for prob in problems:
        pred = agent.predict(prob.train_pairs, prob.test_inputs)
        result.add_answer(prob, pred)

    result.compute_scores()
    return result


if __name__ == "__main__":
    from arc.agents import RandomAgent, CheatingAgent

    agent = RandomAgent()
    results = evaluate_agent(agent)
    print("\nRandom agent:", results)

    agent = CheatingAgent()
    results = evaluate_agent(agent)
    print("\nCheating agent:", results)



from matplotlib.colors import ListedColormap

# RGB
colors_rgb = {
    0: (0x00, 0x00, 0x00),
    1: (0x00, 0x74, 0xD9),
    2: (0xFF, 0x41, 0x36),
    3: (0x2E, 0xCC, 0x40),
    4: (0xFF, 0xDC, 0x00),
    5: (0xA0, 0xA0, 0xA0),
    6: (0xF0, 0x12, 0xBE),
    7: (0xFF, 0x85, 0x1B),
    8: (0x7F, 0xDB, 0xFF),
    9: (0x87, 0x0C, 0x25),
}

_float_colors = [tuple(c / 255 for c in col) for col in colors_rgb.values()]
arc_cmap = ListedColormap(_float_colors)

class ArcColors:
    BLACK = 0
    BLUE = 1
    RED = 2
    GREEN = 3
    YELLOW = 4
    GREY = 5
    FUCHSIA = 6
    ORANGE = 7
    TEAL = 8
    BROWN = 9



import os
import pickle

import appdirs

from arc.version import __version__
from arc.read import parse_dir

here = os.path.dirname(__file__)
train_dir = os.path.join(here, "original", "ARC", "data", "training")
eval_dir = os.path.join(here, "original", "ARC", "data", "evaluation")
cache_dir = appdirs.user_cache_dir("arc-py", "ikamensh")
cache_file = os.path.join(cache_dir, f"arc-py.{__version__}.cache")

if os.path.isfile(cache_file):
    with open(cache_file, "rb") as f:
        train_set, eval_set = pickle.load(f)
else:
    train_set = parse_dir(train_dir)
    eval_set = parse_dir(eval_dir)
    os.makedirs(cache_dir, exist_ok=True)
    with open(cache_file, "wb") as f:
        pickle.dump((train_set, eval_set), f)

assert (
    len(train_set) > 0
), "Training set is empty - this distribution of arc utils is broken."
assert (
    len(eval_set) > 0
), "Evaluation set is empty - this distribution of arc utils is broken."


def describe_task_set(arc_task_set):

    number_examples = [len(x.train_pairs) for x in arc_task_set]
    number_queries = [len(x.test_pairs) for x in arc_task_set]

    from collections import Counter

    print("\nDescribing an ARC task set.")
    print("Number of tasks:", len(arc_task_set))
    print("Number of demonstrations per task:", Counter(number_examples))
    print("Number of tests per task:", Counter(number_queries))


if __name__ == "__main__":

    print("\nTrain")
    describe_task_set(train_set)

    print("\nEval")
    describe_task_set(eval_set)

    for task in train_set:
        for pair in task.train_pairs:
            pair.plot()
            break
        break



from arc.agents import RandomAgent, CheatingAgent
from arc.evaluation import evaluate_agent


def test_random_agent():
    agent = RandomAgent()
    results = evaluate_agent(agent)
    assert results.accuracy < 0.5
    assert results.accuracy_any < 0.5


def test_cheating_agent():
    agent = CheatingAgent()
    results = evaluate_agent(agent)
    assert results.accuracy > 0.99
    assert results.accuracy_any > 0.99
    assert results.shape_accuracy > 0.99



import re
from pathlib import Path

from setuptools import setup

_DIR = Path(__file__).parent


with (_DIR / "arc" / "version.py").open() as f:
    version = re.search('__version__ = "(.*?)"', f.read()).group(1)


def get_requirements():
    with (_DIR / "requirements.txt").open() as f:
        return f.read()


setup(
    name="arc-py",
    packages=["arc"],
    package_dir={"arc": "arc"},
    version=version,
    description="Utilities for working with ARC Challenge.",
    author="Ilya Kamenshchikov",
    keywords=["utility", "ARC"],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
    ],
    license="MIT",
    long_description=(_DIR / "README.md").read_text().strip(),
    long_description_content_type="text/markdown",
    install_requires=get_requirements(),
    python_requires=">=3.6",
    url="https://github.com/ikamensh/arc-py",
    include_package_data=True,
)



import numpy as np

from arc.consts import arc_cmap


def plot_grid(plot_handle, grid: np.ndarray):
    plot_handle.pcolormesh(
        grid,
        cmap=arc_cmap,
        rasterized=True,
        vmin=0,
        vmax=9,
    )
    plot_handle.set_xticks(np.arange(0, grid.shape[1], 1))
    plot_handle.set_yticks(np.arange(0, grid.shape[0], 1))
    plot_handle.grid()
    plot_handle.set_aspect(1)



from typing import List
from abc import ABC

import numpy as np

from arc.data import eval_set
from arc.types import ArcIOPair, ArcGrid, ArcPrediction


class ArcAgent(ABC):
    def predict(
        self, demo_pairs: List[ArcIOPair], test_grids: List[ArcGrid]
    ) -> List[ArcPrediction]:
        """given demonstrations, try to solve test tasks.

        given a list of demo grid pairs for the task,
        output for each test task a sorted list of likely answers.
        The most likely answer is the first in the answer list."""

        raise NotImplementedError


class RandomAgent(ArcAgent):
    """Makes random predicitons. Low chance of success. """

    def predict(
        self, demo_pairs: List[ArcIOPair], test_grids: List[ArcGrid]
    ) -> List[ArcPrediction]:
        outputs = []
        for tg in test_grids:
            out_shape = tg.shape
            out1 = np.random.randint(0, 9, out_shape)
            out2 = np.random.randint(0, 9, out_shape)
            out3 = np.random.randint(0, 9, out_shape)
            outputs.append([out1, out2, out3])
        return outputs


class CheatingAgent(ArcAgent):
    """Cheating agent looks at evaluation set and memorizes answers.

    Expected to get perfect score on the evaluation set."""

    def __init__(self):
        self.answers = {}
        for prob in eval_set:
            for pair in prob.test_pairs:
                self.answers[pair.x.tobytes()] = np.copy(pair.y)

    def predict(
        self, demo_pairs: List[ArcIOPair], test_grids: List[ArcGrid]
    ) -> List[ArcPrediction]:
        outputs = []
        for tg in test_grids:
            outputs.append([self.answers[tg.tobytes()]])
        return outputs


__all__ = [ ArcAgent, RandomAgent, CheatingAgent ]



import os
import json
from typing import List

import numpy as np

from arc.types import ArcIOPair, ArcProblem, ArcGrid, verify_is_arc_grid


def parse_json_grid(grid: List[List[int]]) -> ArcGrid:
    rows = []
    for row in grid:
        rows.append(np.array(row))

    result = np.vstack(rows).astype(np.uint8)
    verify_is_arc_grid(result)
    return result


def parse_dir(d) -> List[ArcProblem]:
    cases = []
    for file in os.listdir(d):
        path = os.path.join(d, file)
        with open(path) as f:
            parsed = json.load(f)

            train = parsed["train"]
            demo_pairs = parse_group(train)

            test = parsed["test"]
            test_pairs = parse_group(test)
            cases.append(
                ArcProblem(
                    demo_pairs=demo_pairs,
                    test_pairs=test_pairs,
                    uid=file.replace(".json", ""),
                )
            )
    return cases


def parse_group(group: List[dict]):
    pairs = []
    for pair in group:
        x = parse_json_grid(pair["input"])
        y = parse_json_grid(pair["output"])
        pairs.append(ArcIOPair(x, y))
    return pairs



import os

import pytest


@pytest.fixture()
def no_cache():
    from arc.data import cache_file
    if os.path.isfile(cache_file):
        os.remove(cache_file)


def test_eval_set(no_cache):
    from arc import eval_set, describe_task_set

    assert len(eval_set) == 400
    describe_task_set(eval_set)


def test_train_set(no_cache):
    from arc import train_set, describe_task_set

    assert len(train_set) == 400
    describe_task_set(train_set)
